package net.shambolica.helperl;

import com.ericsson.otp.erlang.*;

import java.io.UnsupportedEncodingException;
import java.util.Collection;

import static net.shambolica.helperl.EP.pIsAtom;
import static net.shambolica.helperl.EP.pIsBinary;
import static net.shambolica.helperl.EP.pIsList;
import static net.shambolica.helperl.conversion.FromObject.*;
import static net.shambolica.helperl.util.DataUtils.toArray;

/**
 * Convenience constructor functions and predicates for {@link com.ericsson.otp.erlang.OtpErlangObject}s.
 */
public class ET {

    /* =========================================================
     * Construction.
     * ========================================================*/

    /* Default types via erl() */

    public static OtpErlangBinary erl(String string) {
        return eBinary(string);
    }

    public static OtpErlangLong erl(byte n) {
        return eNumber(n);
    }

    public static OtpErlangLong erl(short n) {
        return eNumber(n);
    }

    public static OtpErlangLong erl(int n) {
        return eNumber(n);
    }

    public static OtpErlangLong erl(long n) {
        return eNumber(n);
    }

    public static OtpErlangBinary erl(byte[] bytes) {
        return eBinary(bytes);
    }

    public static OtpErlangList erl(Collection<Object> objects) {
        return eList(objects);
    }


    /* Atoms */

    public static OtpErlangAtom eAtom(Object value) {
        if (value == null) {
            return nullAtom;
        } else if (value instanceof String) {
            return toAtom((String) value);
        } else if (value instanceof Boolean) {
            return ((Boolean)value) ? trueAtom : falseAtom;
        } else if (value instanceof OtpErlangAtom) {
            return (OtpErlangAtom) value;
        } else {
            throw new OEConversionException("unable to convert to atom: "+value.getClass().getName());
        }
    }


    /* Binaries */

    public static OtpErlangBinary eBinary(Object value) {
        if (value == null || (value instanceof byte[])) {
            return eBinary((byte[])value);
        } else if (value instanceof String) {
            return eBinary((String)value);
        } else if (value instanceof OtpErlangBinary) {
            return (OtpErlangBinary) value;
        } else {
            throw new OEConversionException("unable to convert to binary: "+value.getClass().getName());
        }
    }


    /**
     * Create a binary from a byte array.
     */
    public static OtpErlangBinary eBinary(byte... bytes) {
        if (bytes == null || bytes.length == 0) {
            return new OtpErlangBinary(new byte[0]);
        } else {
            return new OtpErlangBinary(bytes);
        }
    }


    /**
     * Convert a string into a UTF-8 binary.
     */
    public static OtpErlangBinary eBinary(String s) {
        if (s.length() == 0) {
            return eBinary();
        } else {
            try {
                byte[] bytes = s.getBytes("UTF-8");
                return eBinary(bytes);
            } catch (UnsupportedEncodingException e) {
                throw new IllegalStateException("UTF-8 not found for binary conversion");
            }
        }
    }


    /* Lists */

    public static OtpErlangList eList(Object ... objects) {
        if (objects == null || objects.length == 0) {
            return new OtpErlangList();
        } else {
            return new OtpErlangList(asObjectsArray(objects));
        }
    }


    public static OtpErlangList eList(Collection<?> objects) {
        return eList(toArray(objects));
    }


    /**
     * Convert a {@link String} into a UTF-8 list.
     */
    public static OtpErlangList eList(String s) {
        return new OtpErlangList(s);
    }


    /* Numbers */

    public static OtpErlangByte eNumber(byte n) {
        return new OtpErlangByte(n);
    }

    public static OtpErlangInt eNumber(int n) {
        return new OtpErlangInt(n);
    }


    public static OtpErlangLong eNumber(long n) {
        return new OtpErlangLong(n);
    }


    public static OtpErlangFloat eNumber(float n) {
        return new OtpErlangFloat(n);
    }


    public static OtpErlangDouble eNumber(double n) {
        return new OtpErlangDouble(n);
    }


    /* Tuples */

    public static OtpErlangTuple eTuple(Object ... objects) {
        if (objects == null || objects.length == 0) {
            return new OtpErlangTuple(new OtpErlangObject[0]);
        } else {
            return new OtpErlangTuple(asObjectsArray(objects));
        }
    }


    public static OtpErlangTuple eTuple(Collection<?> objects) {
        return eTuple(toArray(objects));
    }


    /* Refs */

    public static OtpErlangRef eRef(OtpNode node) {
        return node.createRef();
    }


    public static String getStringValue(OtpErlangObject value) {
        String text;
        if (pIsBinary().matches(value)) {
            try {
                text = new String(((OtpErlangBinary)value).binaryValue(), "UTF-8");
            } catch (UnsupportedEncodingException e) {
                throw new OEConversionException("SHOULD NEVER HAPPEN");
            }
        } else if (pIsAtom().matches(value)) {
            text = ((OtpErlangAtom)value).atomValue();
        } else if (pIsList().matches(value)) {
            try {
                if (value instanceof OtpErlangString) {
                    text = ((OtpErlangString)value).stringValue();
                } else {
                    text = ((OtpErlangList)value).stringValue();
                }
            } catch (OtpErlangException e) {
                throw new OEConversionException("list value does not contain a valid string");
            }
        } else {
            throw new OEConversionException("expected atom|binary|list for text");
        }
        return text;
    }
}
