package net.shambolica.helperl.conversion;

import com.ericsson.otp.erlang.*;
import net.shambolica.helperl.util.DataUtils;

import java.io.UnsupportedEncodingException;

/**
 * Conversion!
 */
public class FromObject {


    /**
     * Returns an array of {@link OtpErlangObject}s representing the input object array,
     * as converted by {@link #asObject(Object)} and padded with "null" atoms using
     * {@link DataUtils#padWithNullAtoms(OtpErlangObject[])}.
     */
    public static OtpErlangObject[] asObjectsArray(Object ... objects) {
        if (objects == null) {
            return null;
        } else {
            OtpErlangObject[] array = new OtpErlangObject[objects.length];
            for (int i = 0 ; i < objects.length ; i++) {
                array[i] = asObject(objects[i]);
            }
            return array;
        }
    }


    /**
     * A version of {@link #asObjectsArray(Object...)} optimized for byte arrays.
     */
    public static OtpErlangLong[] asObjectsArray(byte ... bytes) {
        OtpErlangLong[] objects = new OtpErlangLong[bytes.length];
        for (int i = 0 ; i < bytes.length ; i++) {
            byte b = bytes[i];
            objects[i] = byteConstant(b);
        }
        return objects;
    }


    /**
     * Convert basic data types to {@link OtpErlangObject}s.
     * Nulls will be converted to {@link #nullAtom}.
     */
    public static OtpErlangObject asObject(Object object) {
        if (object == null) {
            return nullAtom;
        } else if (object instanceof OtpErlangObject) {
            return (OtpErlangObject) object;
        } else if (object instanceof Byte) {
            return byteConstant((Byte)object);
        } else if (object instanceof Short) {
            return longErlangObject((Short) object);
        } else if (object instanceof Integer) {
            return longErlangObject((Integer) object);
        } else if (object instanceof Long) {
            return longErlangObject((Long) object);
        } else if (object instanceof Float) {
            return doubleErlangObject((Float)object);
        } else if (object instanceof Double) {
            return doubleErlangObject((Double)object);
        } else if (object instanceof byte[]) {
            return new OtpErlangBinary((byte[])object);
        } else if (object instanceof String) {
            try {
                return new OtpErlangBinary(((String)object).getBytes("UTF-8"));
            } catch (UnsupportedEncodingException e) {
                throw new IllegalStateException("cannot find UTF-8");
            }
        } else {
            throw new IllegalArgumentException("unexpected call to asObject with "+object.getClass().getName());
        }
    }


    /**
     * Convert a {@link String} to an atom, using the constant atoms if possible.
     * Note that a null will be converted into {@link #nullAtom}.
     */
    public static OtpErlangAtom toAtom(String value) {
        if (value == null) {
            return nullAtom;
        } else if (value.equals("true")) {
            return trueAtom;
        } else if (value.equals("false")) {
            return falseAtom;
        } else if (value.equals("undefined")) {
            return undefinedAtom;
        } else if (value.equals("null")) {
            return nullAtom;
        } else {
            return new OtpErlangAtom(value);
        }
    }


    /* =========================================================
     * Constants.
     * ========================================================*/

    /* Atoms */

    public static final OtpErlangAtom nullAtom = new OtpErlangAtom("null");
    public static final OtpErlangAtom undefinedAtom = new OtpErlangAtom("undefined");
    public static final OtpErlangBoolean trueAtom = new OtpErlangBoolean(true);
    public static final OtpErlangBoolean falseAtom = new OtpErlangBoolean(false);


    /* Bytes */

    private static OtpErlangLong[] byteValues = new OtpErlangLong[256];
    static {
        for (int i = 0 ; i < 256 ; i++) {
            byteValues[i] = new OtpErlangLong(i);
        }
    }

    private static OtpErlangLong[] negativeByteValues = new OtpErlangLong[256];
    static {
        for (int i = 0 ; i < 256 ; i++) {
            negativeByteValues[i] = new OtpErlangLong(-i);
        }
    }


    /* Floats */

    public static OtpErlangDouble zeroDouble = new OtpErlangDouble(0.0);
    public static OtpErlangDouble plus1Double = new OtpErlangDouble(1.0);
    public static OtpErlangDouble plus2Double = new OtpErlangDouble(2.0);
    public static OtpErlangDouble minus1Double = new OtpErlangDouble(-1.0);

    /* =========================================================
     * Utilities.
     * ========================================================*/


    public static OtpErlangLong byteConstant(byte b) {
        return b < 0 ? negativeByteValues[-b] : byteValues[b];
    }

    public static OtpErlangLong longErlangObject(long n) {
        if (n == (n & 0xff)) {
            return byteConstant((byte)n);
        } else {
            return new OtpErlangLong(n);
        }
    }

    public static OtpErlangDouble doubleErlangObject(double n) {
        if (n == 0.0) {
            return zeroDouble;
        } else if (n == -1.0) {
            return minus1Double;
        } else if (n == 1.0) {
            return plus1Double;
        } else if (n == 2.0) {
            return plus2Double;
        } else {
            return new OtpErlangDouble(n);
        }
    }

}
